@@ -30,6 +30,7 @@ from .utils import (
     _connect_engine,
     _get_heads,
     _repair_self_loop_if_present,
+    _resolve_down_revision,
 )
 
 app = typer.Typer(
@@ -200,200 +201,196 @@ format = %(levelname)-5.5s [%(name)s] %(message)s
 
     # 2) env.py — sync only, async→sync mapping is handled at runtime
     env_py = project_root / AL_EMBIC_DIR / "env.py"
-    if not env_py.exists():
-        content = dedent(
-            f"""\
-            from __future__ import annotations
-            import os
-            import sys
-            import logging
-            import pkgutil
-            import importlib
-            from pathlib import Path
-            from logging.config import fileConfig
-            from typing import Iterable, List
-
-            from alembic import context
-            from sqlalchemy import pool, create_engine
-            from sqlalchemy.engine.url import make_url, URL
-            from sqlalchemy.orm import DeclarativeBase
-
-            # --- Ensure project root and src/ on sys.path ---
-            ROOT = Path(__file__).resolve().parents[1]
-            for p in (ROOT, ROOT / "src"):
-                s = str(p)
-                if p.exists() and s not in sys.path:
-                    sys.path.insert(0, s)
-
-            # --- App logging (optional) ---
-            USE_APP_LOGGING = os.getenv("ALEMBIC_USE_APP_LOGGING", "1") == "1"
-            if USE_APP_LOGGING:
+    # Always (re)write env.py to ensure it's correct; if you prefer idempotency,
+    # change to "if not env_py.exists():" again after you’re confident.
+    # For now, be robust for the agent:
+    # if not env_py.exists():
+    env_tpl = """
+    from __future__ import annotations
+    import os
+    import sys
+    import logging
+    import importlib
+    from pathlib import Path
+    from logging.config import fileConfig
+    from typing import Iterable, List
+    
+    from alembic import context
+    from sqlalchemy import pool, create_engine
+    from sqlalchemy.engine.url import make_url, URL
+    from sqlalchemy.orm import DeclarativeBase
+    
+    # --- Ensure project root and src/ on sys.path ---
+    ROOT = Path(__file__).resolve().parents[1]
+    for p in (ROOT, ROOT / "src"):
+        s = str(p)
+        if p.exists() and s not in sys.path:
+            sys.path.insert(0, s)
+    
+    # --- Optional: load .env for Alembic runtime (noop if missing) ---
+    try:
+        from dotenv import load_dotenv  # type: ignore
+        load_dotenv(dotenv_path=ROOT / ".env")
+    except Exception:
+        pass
+    
+    # --- Alembic config & logging ---
+    config = context.config
+    USE_APP_LOGGING = os.getenv("ALEMBIC_USE_APP_LOGGING", "1") == "1"
+    if USE_APP_LOGGING:
+        try:
+            from svc_infra.app.logging import setup_logging
+            setup_logging(level=os.getenv("LOG_LEVEL"), fmt=os.getenv("LOG_FORMAT"))
+            logging.getLogger(__name__).debug("Alembic using app logging setup.")
+        except Exception as e:
+            USE_APP_LOGGING = False
+            print(f"[alembic] App logging import failed: {e}. Falling back to fileConfig.")
+    
+    if not USE_APP_LOGGING and config.config_file_name is not None:
+        fileConfig(config.config_file_name)
+        logging.getLogger(__name__).debug("Alembic using fileConfig logging.")
+    
+    # --- DATABASE_URL override via env (if ini is blank/placeholder) ---
+    env_url = os.getenv("DATABASE_URL")
+    if env_url:
+        config.set_main_option("sqlalchemy.url", env_url)
+    
+    # --- Import ONLY model modules to avoid app startup side-effects ---
+    # Provide packages via ALEMBIC_DISCOVER_PACKAGES: "pkg1,pkg2"
+    DISCOVER_PKGS = os.getenv("ALEMBIC_DISCOVER_PACKAGES", "__DISCOVER__")
+    
+    def import_model_modules(packages: Iterable[str]) -> None:
+        # Only try <pkg>.models and <pkg>.db.models
+        candidates = ("models", "db.models")
+        for pkg in packages:
+            pkg = (pkg or "").strip()
+            if not pkg:
+                continue
+            for leaf in candidates:
+                mod = f"{pkg}.{leaf}"
                 try:
-                    from svc_infra.app.logging import setup_logging
-                    setup_logging(level=os.getenv("LOG_LEVEL"), fmt=os.getenv("LOG_FORMAT"))
-                    logging.getLogger(__name__).debug("Alembic using app logging setup.")
+                    importlib.import_module(mod)
                 except Exception as e:
-                    USE_APP_LOGGING = False
-                    print(f"[alembic] App logging import failed: {{e}}. Falling back to fileConfig.")
-
-            # --- Alembic config & logging ---
-            config = context.config
-            if not USE_APP_LOGGING and config.config_file_name is not None:
-                fileConfig(config.config_file_name)
-                logging.getLogger(__name__).debug("Alembic using fileConfig logging.")
-
-            # --- Database URL override via env (if ini is blank) ---
-            env_url = os.getenv("DATABASE_URL")
-            if env_url:
-                config.set_main_option("sqlalchemy.url", env_url)
-
-            # --- Auto-discover model modules and collect all metadatas ---
-            DISCOVER_PKGS = os.getenv("ALEMBIC_DISCOVER_PACKAGES", "{discover_root_csv}")
-
-            def _iter_pkg_modules(top_pkg_name: str) -> Iterable[str]:
-                try:
-                    top_pkg = importlib.import_module(top_pkg_name)
-                except Exception:
-                    return []
-                if not hasattr(top_pkg, "__path__"):
-                    return [top_pkg_name]
-                names = []
-                for m in pkgutil.walk_packages(top_pkg.__path__, prefix=top_pkg.__name__ + "."):
-                    names.append(m.name)
-                return names
-
-            def import_all_under_packages(packages: Iterable[str]) -> None:
-                for pkg_name in packages:
-                    if not pkg_name:
-                        continue
-                    for mod_name in _iter_pkg_modules(pkg_name):
-                        try:
-                            importlib.import_module(mod_name)
-                        except Exception as e:
-                            logging.getLogger(__name__).debug(f"[alembic] Skipped import {{mod_name}}: {{e}}")
-
-            def collect_all_metadatas() -> List:
-                metas: set = set()
-                try:
-                    for cls in DeclarativeBase.__subclasses__():
-                        md = getattr(cls, "metadata", None)
-                        if md is not None:
-                            metas.add(md)
-                except Exception:
-                    pass
-                return list(metas)
-
-            pkgs = [p.strip() for p in (DISCOVER_PKGS or "").split(",") if p.strip()]
-            import_all_under_packages(pkgs)
-            target_metadata = collect_all_metadatas()  # may be empty (autogen no-op)
-
-            # --- URL normalization: map async drivers to sync + ensure sslmode=require ---
-            def _sync_url_for_migrations(url_str: str) -> str:
-                if not url_str:
-                    return url_str
-                try:
-                    u: URL = make_url(url_str)
-                except Exception:
-                    return url_str
-
-                drv = (u.drivername or "")
-                # map async → sync driver
-                if drv.endswith("+asyncpg"):
-                    u = u.set(drivername="postgresql+psycopg2")
-                elif drv.endswith("+aiosqlite"):
-                    u = u.set(drivername="sqlite")
-                elif drv.endswith("+asyncmy"):
-                    u = u.set(drivername="mysql+pymysql")
-
-                # ensure sslmode=require for postgres if missing
-                try:
-                    from urllib.parse import urlencode
-                    if u.drivername.startswith("postgresql"):
-                        q = dict(u.query)
-                        if "sslmode" not in q:
-                            q["sslmode"] = "require"
-                        u = u.set(query=q)
-                except Exception:
-                    pass
-
-                return str(u)
-
-            def run_migrations_offline():
-                url = _sync_url_for_migrations(config.get_main_option("sqlalchemy.url"))
-                context.configure(
-                    url=url,
-                    target_metadata=target_metadata,
-                    literal_binds=True,
-                    compare_type=True,
-                    compare_server_default=True,
-                    render_as_batch=True,
-                )
-                with context.begin_transaction():
-                    context.run_migrations()
-
-            def do_run_migrations(connection):
-                context.configure(
-                    connection=connection,
-                    target_metadata=target_metadata,
-                    compare_type=True,
-                    compare_server_default=True,
-                    render_as_batch=True,
-                )
-                with context.begin_transaction():
-                    context.run_migrations()
-
-            def run_migrations_online_sync():
-                url = _sync_url_for_migrations(config.get_main_option("sqlalchemy.url"))
-                engine = create_engine(url, poolclass=pool.NullPool, future=True)
-                try:
-                    with engine.connect() as connection:
-                        do_run_migrations(connection)
-                finally:
-                    engine.dispose()
-
-            if context.is_offline_mode():
-                run_migrations_offline()
-            else:
-                run_migrations_online_sync()
-            """
+                    logging.getLogger(__name__).debug(f"[alembic] Skipped import {mod}: {e}")
+    
+    def collect_all_metadatas() -> List:
+        metas: set = set()
+        try:
+            for cls in DeclarativeBase.__subclasses__():
+                md = getattr(cls, "metadata", None)
+                if md is not None:
+                    metas.add(md)
+        except Exception:
+            pass
+        return list(metas)
+    
+    pkgs = [p.strip() for p in (DISCOVER_PKGS or "").split(",") if p.strip()]
+    import_model_modules(pkgs)
+    target_metadata = collect_all_metadatas()  # may be empty (autogen no-op)
+    
+    # --- URL normalization: map async → sync; require SSL for Postgres ---
+    def _sync_url_for_migrations(url_str: str) -> str:
+        if not url_str:
+            return url_str
+        try:
+            u: URL = make_url(url_str)
+        except Exception:
+            return url_str
+    
+        drv = (u.drivername or "")
+        if drv.endswith("+asyncpg"):
+            u = u.set(drivername="postgresql+psycopg2")
+        elif drv.endswith("+aiosqlite"):
+            u = u.set(drivername="sqlite")
+        elif drv.endswith("+asyncmy"):
+            u = u.set(drivername="mysql+pymysql")
+    
+        if u.drivername.startswith("postgresql"):
+            q = dict(u.query)
+            if "sslmode" not in q:
+                q["sslmode"] = "require"
+            u = u.set(query=q)
+    
+        return str(u)
+    
+    def run_migrations_offline():
+        url = _sync_url_for_migrations(config.get_main_option("sqlalchemy.url"))
+        context.configure(
+            url=url,
+            target_metadata=target_metadata,
+            literal_binds=True,
+            compare_type=True,
+            compare_server_default=True,
+            render_as_batch=True,
         )
-        env_py.write_text(content, encoding="utf-8")
-        created["env_py"] = True
-        notes.append(f"Wrote {env_py}")
+        with context.begin_transaction():
+            context.run_migrations()
+    
+    def do_run_migrations(connection):
+        context.configure(
+            connection=connection,
+            target_metadata=target_metadata,
+            compare_type=True,
+            compare_server_default=True,
+            render_as_batch=True,
+        )
+        with context.begin_transaction():
+            context.run_migrations()
+    
+    def run_migrations_online_sync():
+        url = _sync_url_for_migrations(config.get_main_option("sqlalchemy.url"))
+        engine = create_engine(url, poolclass=pool.NullPool, future=True)
+        try:
+            with engine.connect() as connection:
+                do_run_migrations(connection)
+        finally:
+            engine.dispose()
+    
+    if context.is_offline_mode():
+        run_migrations_offline()
+    else:
+        run_migrations_online_sync()
+    """
+    # Fill in the discover CSV, then dedent and strip leading blank lines
+    env_text = dedent(env_tpl.replace("__DISCOVER__", discover_root_csv)).lstrip()
+    env_py.write_text(env_text + "\n", encoding="utf-8")
 
-    # 2.5) script.py.mako (unchanged)
-    script_tpl = dedent(
-        """\
+    # Fail fast if we somehow wrote a bad file
+    py_compile.compile(str(env_py), doraise=True)
+    created["env_py"] = True
+    notes.append(f"Wrote {env_py}")
+    script_path = project_root / AL_EMBIC_DIR / "script.py.mako"
+    if not script_path.exists():
+        script_tpl = """
         \"\"\"${message}
-
+        
         Revision ID: ${up_revision}
         Revises: ${down_revision | comma,n}
         Create Date: ${create_date}
-
+        
         \"\"\"
         from __future__ import annotations
-
+        
         from alembic import op
         import sqlalchemy as sa
         import fastapi_users_db_sqlalchemy  # noqa: F401
-
+        
         revision: str = ${repr(up_revision)}
         down_revision: str | None = ${repr(down_revision)}
         branch_labels: tuple[str, ...] | None = ${repr(branch_labels)}
         depends_on: tuple[str, ...] | None = ${repr(depends_on)}
-
+        
         def upgrade() -> None:
             ${upgrades if upgrades else "pass"}
-
+        
         def downgrade() -> None:
             ${downgrades if downgrades else "pass"}
         """
-    )
-    script_path = project_root / AL_EMBIC_DIR / "script.py.mako"
-    if not script_path.exists():
-        script_path.write_text(script_tpl, encoding="utf-8")
-        created["script_tpl"] = True
-        notes.append(f"Wrote {script_path}")
+        script_path = project_root / AL_EMBIC_DIR / "script.py.mako"
+        if not script_path.exists():
+            script_path.write_text(script_tpl, encoding="utf-8")
+            created["script_tpl"] = True
+            notes.append(f"Wrote {script_path}")
 
     versions = _versions_dir(project_root)
     versions.mkdir(exist_ok=True)
@@ -413,17 +410,21 @@ format = %(levelname)-5.5s [%(name)s] %(message)s
         "notes": notes,
     }
 
-def db_revision_core(
-        *,
-        message: str,
-        autogenerate: bool,
-        project_root: Path,
-        database_url: Optional[str] = None,
-) -> Dict[str, Any]:
-    """Create a new Alembic revision."""
+def db_revision_core(*, message: str, autogenerate: bool, project_root: Path, database_url: Optional[str] = None) -> Dict[str, Any]:
     project_root = project_root.resolve()
     _load_dotenv_if_present(project_root)
     cfg = _load_config(project_root, database_url)
+
+    eff = cfg.get_main_option("sqlalchemy.url")
+    if eff:
+        try:
+            engine = create_engine(eff, future=True)
+            with closing(engine.connect()) as conn:
+                conn.exec_driver_sql("SELECT 1")
+            engine.dispose()
+        except Exception as e:
+            raise RuntimeError(f"DB connection check failed before autogenerate: {e}") from e
+
     command.revision(cfg, message=message, autogenerate=autogenerate)
     return {"status": "ok", "message": message, "autogenerate": autogenerate}
 
@@ -497,125 +498,93 @@ def db_drop_table_core(
         base: Optional[str],
         apply: bool,
         project_root: Path,
-        database_url: Optional[str] = None,
+        database_url: Optional[str],
 ) -> Dict[str, Any]:
     """
     Create and optionally apply a migration that drops a specified table.
-    - Ensures Alembic is bootstrapped
-    - Pings DB first
-    - Creates a revision whose down_revision is computed from heads *before* creation
-    - Only rewrites function bodies; preserves header generated by Alembic
-    - Applies upgrade if requested
-    - Verifies table removal
+    Safe against Alembic self-loops; supports base=None/"head"/<rev_id>.
     """
     project_root = project_root.resolve()
 
-    # 0) Ensure alembic env exists
+    # 0) Ensure Alembic is bootstrapped (idempotent)
     _ensure_alembic_bootstrap(project_root=project_root, database_url=database_url)
 
-    # 0.1) If repo is already broken by a self-loop, repair it so we can proceed
-    _repair_self_loop_if_present(project_root)
-
-    # 1) Effective DB url
+    # 1) Effective DB url & ping
     eff_url = _effective_migration_url(project_root, database_url)
-
-    # 2) Quick ping & pre-state
     engine = _connect_engine(eff_url)
     try:
         with engine.connect() as conn:
-            conn.execute(text("SELECT 1"))
+            conn.exec_driver_sql("SELECT 1")  # SQLA 2.0-safe ping
     except Exception as e:
         engine.dispose()
         raise RuntimeError(f"DB ping failed: {e}") from e
 
+    # Determine schema/table existence BEFORE
     schema, tbl = _parse_table_identifier(table)
     existed_before = _table_exists(engine, schema, tbl)
 
-    # 3) Compute down_revision *before* creating a new revision
+    # 2) Prepare Alembic config & resolve down_revision *BEFORE* creating the new revision
     cfg = _load_config(project_root, database_url)
     script = _script_dir(cfg)
-    heads_before = _get_heads(script)
 
-    if base:
-        down_rev_to_use = base
-    else:
-        if len(heads_before) == 0:
-            down_rev_to_use = None
-        elif len(heads_before) == 1:
-            down_rev_to_use = heads_before[0]
-        else:
-            raise RuntimeError(
-                "Multiple heads present. Provide base=<rev> or merge heads first."
-            )
+    try:
+        down_rev = _resolve_down_revision(script, base)
+    except RuntimeError as e:
+        engine.dispose()
+        # Preserve the helpful message about multiple heads
+        raise
 
-    # 4) Create a new revision file
+    # 3) Create an empty revision file (autogenerate=False)
     msg = message or f"drop table {table}"
     command.revision(cfg, message=msg, autogenerate=False)
 
+    # 4) Open the newest file and write our content with the resolved down_revision
     versions_dir = project_root / AL_EMBIC_DIR / "versions"
     rev_file = _latest_version_file(versions_dir)
     if rev_file is None:
         engine.dispose()
         raise RuntimeError("Could not locate newly created revision file.")
 
-    # 5) Rewrite only header's down_revision (if needed) and the bodies
-    txt = rev_file.read_text(encoding="utf-8")
+    rev_id = rev_file.stem.split("_", 1)[0]
 
-    # Replace down_revision line to the correct prior head (or None) if present
-    # Keep the Alembic-generated 'revision = ...' untouched.
-    if down_rev_to_use is None:
-        txt = re.sub(r'^\s*down_revision\s*=.*$', 'down_revision = None', txt, flags=re.M)
-    else:
-        txt = re.sub(
-            r'^\s*down_revision\s*=.*$',
-            f'down_revision = "{down_rev_to_use}"',
-            txt,
-            flags=re.M,
-        )
-
-    # Generate safe SQL
-    qschema = _quote_ident(schema) + "." if schema else ""
+    # Build safe SQL
+    qschema = (_quote_ident(schema) + ".") if schema else ""
     qtable = _quote_ident(tbl)
     sql = f'DROP TABLE {"IF EXISTS " if if_exists else ""}{qschema}{qtable}{" CASCADE" if cascade else ""};'
 
-    # Replace/insert upgrade() body
-    txt = re.sub(
-        r"def upgrade\(\)\s*->\s*None:\n(\s+.*\n)*",
-        f"def upgrade() -> None:\n    op.execute({sql!r})\n\n",
-        txt,
-        flags=re.M,
+    if down_rev == rev_id:
+        # This should never happen now; bail loudly if it does.
+        raise RuntimeError(
+            f"Refusing to write self-referencing revision: down_revision == revision ({rev_id})."
+        )
+
+    content = (
+        f'''""" {msg} """\n\n'''
+        f'''from __future__ import annotations\n\n'''
+        f'''from alembic import op\nimport sqlalchemy as sa\n\n'''
+        f'''revision = "{rev_id}"\n'''
+        f'''down_revision = {repr(down_rev) if down_rev else "None"}\n'''
+        f'''branch_labels = None\n'''
+        f'''depends_on = None\n\n'''
+        f'''def upgrade() -> None:\n'''
+        f'''    op.execute({sql!r})\n\n'''
+        f'''def downgrade() -> None:\n'''
+        f'''    # Irreversible without full table definition\n'''
+        f'''    pass\n'''
     )
+    rev_file.write_text(content, encoding="utf-8")
 
-    # Replace/insert downgrade() body (irreversible)
-    txt = re.sub(
-        r"def downgrade\(\)\s*->\s*None:\n(\s+.*\n)*",
-        "def downgrade() -> None:\n    # Irreversible without full table definition\n    pass\n",
-        txt,
-        flags=re.M,
-    )
-
-    rev_file.write_text(txt, encoding="utf-8")
-
-    # 6) Apply if requested
+    # 5) Apply if requested
     applied = False
     if apply:
         try:
             command.upgrade(cfg, "head")
             applied = True
-        except Exception as e:
+        except Exception:
             engine.dispose()
-            # If we somehow created a self-loop anyway, attempt one-shot repair and retry once
-            if "Self-loop is detected" in str(e):
-                fix = _repair_self_loop_if_present(project_root)
-                if fix.get("repaired"):
-                    command.upgrade(cfg, "head")
-                    applied = True
-                else:
-                    raise
-            else:
-                raise
+            raise
 
-    # 7) Verify AFTER
+    # 6) Verify AFTER
     exists_after = _table_exists(engine, schema, tbl)
     engine.dispose()
 
@@ -628,6 +597,7 @@ def db_drop_table_core(
         "if_exists": if_exists,
         "existed_before": bool(existed_before),
         "exists_after": bool(exists_after),
+        "note": "If exists_after is True, your DB user may lack privileges or the table is in a different schema.",
     }
 
 def db_merge_heads_core(*, message: str, project_root: Path, database_url: Optional[str] = None) -> Dict[str, Any]:
